@@ -1,5 +1,5 @@
 (* 
- * $Id: unixqueue.ml 45 2004-10-07 15:45:09Z gerd $
+ * $Id: unixqueue.ml 56 2005-09-03 14:12:12Z gerd $
  *)
 
 (**********************************************************************)
@@ -227,6 +227,10 @@ type handler =
  * to the watched resources such that when bytes arrive thread B wakes
  * up. Thread A signals changed conditions by writing a single byte into
  * the write end of the control pipe.
+ *
+ * The control pipe is now created every time [run] is called, and it is
+ * closed before this method is left. This avoids that we run out of 
+ * file descriptors when lots of queues are created. 
  *)
 
 exception Abort of (group * exn)
@@ -241,14 +245,6 @@ exception Exit;;   (* used locally in uq_handler *)
 
 class unix_event_system () : event_system =
   let es_lock, es_unlock = !create_lock_unlock_pair() in
-  let ctrl_rd, ctrl_wr =
-    if !mt_mode then
-      let rd,wr = pipe() in
-      set_nonblock rd;
-      rd,wr
-    else
-      stdin,stdin in
-  let ctrl_g = new group_object in
 object(self)
 
   val mutable sys = lazy (assert false)   (* initialized below *)
@@ -265,9 +261,9 @@ object(self)
   val mutable lock = es_lock
   val mutable unlock = es_unlock
   val mutable blocking = false
-  val mutable ctrl_pipe_rd = ctrl_rd
-  val mutable ctrl_pipe_wr = ctrl_wr
-  val mutable ctrl_pipe_group = ctrl_g
+  val mutable ctrl_pipe_rd = None
+  val mutable ctrl_pipe_wr = None
+  val mutable ctrl_pipe_group = new group_object
 
   (**********************************************************************)
   (* Implementation of the queue                                        *)
@@ -277,9 +273,6 @@ object(self)
     let equeue_sys = Equeue.create ~string_of_event self#source in
     sys <- lazy equeue_sys;
     ignore(Lazy.force sys);
-    (* In multi-threaded mode: watch at least the control pipe *)
-    if !mt_mode then
-      res <- RID_Map.add (Wait_in ctrl_rd) (ctrl_g, -1.0, -1.0) RID_Map.empty;
 
   method private setup () =
     (* Find out which resources should be watched for interesting events *)
@@ -394,20 +387,25 @@ object(self)
        * correctly.
        * Update: Does not work. Don't know how to fix this problem.
        *)
-      try
-	while true do
-	  let have_ctrl_data, _, _ = select [ ctrl_pipe_rd ] [] [] 0.0 in
-	  if have_ctrl_data = [] then
-	    raise(Unix_error(EAGAIN,"(artificial)",""));
-	  ignore(read ctrl_pipe_rd dummy_buf 0 1)
-	done;
-	assert false
-      with
-	  Unix_error(EAGAIN,_,_) ->
-	    ()
-	| Unix_error(EINTR,_,_) as err ->
-	    (* It is unclear whether this can happen or not. *)
-	    deferred_exn := Some err
+      match ctrl_pipe_rd with
+	| None ->
+	    ()   (* this should not happen *)
+	| Some p ->
+	    ( try
+		while true do
+		  let have_ctrl_data, _, _ = select [ p ] [] [] 0.0 in
+		  if have_ctrl_data = [] then
+		    raise(Unix_error(EAGAIN,"(artificial)",""));
+		  ignore(read p dummy_buf 0 1)
+		done;
+		assert false
+	      with
+		  Unix_error(EAGAIN,_,_) ->
+		    ()
+		| Unix_error(EINTR,_,_) as err ->
+		    (* It is unclear whether this can happen or not. *)
+		    deferred_exn := Some err
+	    )
     in
 
     (* first process file descriptors, and add events resulting from 
@@ -495,7 +493,6 @@ object(self)
     lock();
     try
       (* Find out which system events are interesting now: *)
-	
       let (infiles, outfiles, oobfiles, time) as watch_tuple = self#setup() in
 
       debug_print
@@ -508,7 +505,7 @@ object(self)
 	   ));
 
       let interesting =
-	rid_map_exists (fun _ (g,_,_) -> g <> ctrl_g) res in
+	rid_map_exists (fun _ (g,_,_) -> g <> ctrl_pipe_group) res in
 
       if interesting then begin
 	(* There ARE interesting situations. *)
@@ -549,7 +546,9 @@ object(self)
       end;
       unlock()
     with
-	any -> unlock(); raise any
+	any -> 
+	  unlock(); 
+	  raise any
 
   (**********************************************************************)
   (* External interface                                                 *)
@@ -612,8 +611,12 @@ object(self)
 	(* avoid that the queue becomes empty: *)
 	if keep_alive then 
 	  Equeue.add_event (Lazy.force sys) (Extra Keep_alive); 
-	let buf = String.make 1 'X' in
-	ignore(write ctrl_pipe_wr buf 0 1);
+	( match ctrl_pipe_wr with
+	    | None -> ()
+	    | Some p_wr ->
+		let buf = String.make 1 'X' in
+		ignore(single_write p_wr buf 0 1);
+	)
       with
 	  Unix_error(EINTR,_,_) ->
 	    Equeue.add_event (Lazy.force sys) Signal;
@@ -943,26 +946,58 @@ object(self)
   method private abort g =
   self#protect self#abort_nolock g
 
+
+  method private add_ctrl_pipe() =
+    (* In multi-threaded mode: watch at least the control pipe *)
+    if !mt_mode then (
+      assert(ctrl_pipe_rd = None && ctrl_pipe_wr = None);
+      let (p_rd, p_wr) = Unix.pipe() in
+      ctrl_pipe_rd <- Some p_rd;
+      ctrl_pipe_wr <- Some p_wr;
+      new_res <- RID_Map.add (Wait_in p_rd) (ctrl_pipe_group,-1.0,-1.0) new_res
+    );
+
+  method private close_ctrl_pipe() =
+    match ctrl_pipe_rd, ctrl_pipe_wr with
+      | None, None -> ()
+      | (Some p_rd), (Some p_wr) ->
+	  let op = Wait_in p_rd in
+	  res     <- RID_Map.remove op res;
+	  new_res <- RID_Map.remove op new_res;
+	  Unix.close p_wr;
+	  Unix.close p_rd;
+	  ctrl_pipe_rd <- None;
+	  ctrl_pipe_wr <- None;
+      | _ ->
+	  assert false
+
   method run () =
     let continue = ref true in
-    while !continue do
-      continue := false;
-      try
-	Equeue.run (Lazy.force sys);
-      with
-	  Abort (g,an_exception) ->
-	    begin
-	      match an_exception with
-		  (Equeue.Reject|Equeue.Terminate) ->
-		    (* A serious programming error: *)
-		    failwith "Caught 'Abort' exception with Reject or Terminate exception as argument; this is a programming error"
-		| Abort(_,_) ->
-		    failwith "Caught 'Abort' exception with an 'Abort' exception as argument; this is a programming error"
-		| _ -> ()
-	    end;
-	    self#abort g an_exception;
-	    continue := true
-    done
+    self # add_ctrl_pipe();
+    try
+      while !continue do
+	continue := false;
+	try
+	  Equeue.run (Lazy.force sys);
+	with
+	    Abort (g,an_exception) ->
+	      begin
+		match an_exception with
+		    (Equeue.Reject|Equeue.Terminate) ->
+		      (* A serious programming error: *)
+		      failwith "Caught 'Abort' exception with Reject or Terminate exception as argument; this is a programming error"
+		  | Abort(_,_) ->
+		      failwith "Caught 'Abort' exception with an 'Abort' exception as argument; this is a programming error"
+		  | _ -> ()
+	      end;
+	      self#abort g an_exception;
+	      continue := true
+      done;
+      self # close_ctrl_pipe();
+    with
+      | error ->
+	  self # close_ctrl_pipe();
+	  raise error
 
   method once g duration f =
     let id = self#new_wait_id () in
