@@ -1,5 +1,5 @@
 (* 
- * $Id: uq_engines.ml 45 2004-10-07 15:45:09Z gerd $
+ * $Id: uq_engines.ml 53 2005-05-14 11:35:45Z gerd $
  *)
 
 
@@ -567,6 +567,7 @@ object(self)
 
   val mutable out_eof = false
 
+  val mutable deferred_exn = None
 
   initializer
     (* Arrange that Receiver_attn is generated when the dst state changes: *)
@@ -671,8 +672,14 @@ object(self)
 	    (* These exceptions are expected, and can be ignored *)
 	    ()
 	| error ->
-	    (* Any other exception stops the engine *)
-	    raise(Unixqueue.Abort(group,error))
+	    (* Any other exception stops the engine. But first it is tried to
+	     * process the buffer contents:
+	     *)
+	    in_eof <- true;
+	    deferred_exn <- Some error;
+	    if in_eof && close_src then Unix.close src;
+	    Unixqueue.add_event ues (receiver_attn group);
+	    self # count();
 	    
 
   method private check_input_polling() =
@@ -724,7 +731,10 @@ object(self)
 	     *)
 	    if close_dst then dst # close_out();
 	    out_eof <- true;
-	    self # set_state (`Done());
+	    ( match deferred_exn with
+		| None -> self # set_state (`Done());
+		| Some err -> self # set_state (`Error err);
+	    )
 	  )
 	)
       with
@@ -861,9 +871,9 @@ object(self)
 
 
   method private handle_output() =
-    if not out_eof && buf_size > 0 then
+    if not out_eof then
       try
-	let n = Unix.write dst buf 0 buf_size in
+	let n = Unix.single_write dst buf 0 buf_size in
 	String.blit buf n buf 0 (buf_size - n);
 	buf_size <- buf_size - n;
 	if buf_size = 0 && in_eof then (
@@ -1128,7 +1138,7 @@ object (self)
       (* Write buf to dst: *)
       try
 	let n =
-	  if buf_size > 0 then Unix.write dst buf 0 buf_size else 0 in
+	  if buf_size > 0 then Unix.single_write dst buf 0 buf_size else 0 in
 	String.blit buf n buf 0 (buf_size - n);
 	buf_size <- buf_size - n;
 	if in_eof && buf_size = 0 then (
@@ -1568,7 +1578,13 @@ object (self)
 		  setup_socket s stype (Unix.ADDR_INET(addr,port)) opts;
 	  in
 	  if is_connected then (
-	    new const_engine (`Done(`Socket(s, getsockspec stype s))) ues
+	    let status =
+	      try 
+		ignore(Unix.getpeername s); 
+		`Done(`Socket(s, getsockspec stype s))
+	      with
+		  error -> `Error error in
+	    new const_engine status ues
 	  )
 	  else (
 	    (* Now wait until the socket is writeable *)
@@ -1702,6 +1718,7 @@ object(self)
       new map_engine 
 	~map_done:(fun _ ->
 		     let (sock',_) = Unix.accept sock in
+		     Unix.set_nonblock sock';
 		     acc_engine <- None;
 		     `Done(sock', getpeerspec Unix.SOCK_STREAM sock')
 		  )
@@ -1737,6 +1754,7 @@ object(self)
 	    ( match sockspec with
 		  `Sock_unix(stype, path) ->
 		    let s = Unix.socket Unix.PF_UNIX stype 0 in
+		    Unix.set_nonblock s;
 		    if opts.lstn_reuseaddr then 
 		      Unix.setsockopt s Unix.SO_REUSEADDR true;
 		    Unix.bind s (Unix.ADDR_UNIX path);
@@ -1744,6 +1762,7 @@ object(self)
 		    s
 		| `Sock_inet(stype, addr, port) ->
 		    let s = Unix.socket Unix.PF_INET stype 0 in
+		    Unix.set_nonblock s;
 		    if opts.lstn_reuseaddr then 
 		      Unix.setsockopt s Unix.SO_REUSEADDR true;
 		    Unix.bind s (Unix.ADDR_INET(addr,port));
@@ -1752,6 +1771,7 @@ object(self)
 		| `Sock_inet_byname(stype, name, port) ->
 		    let addr = addr_of_name name in
 		    let s = Unix.socket Unix.PF_INET stype 0 in
+		    Unix.set_nonblock s;
 		    if opts.lstn_reuseaddr then 
 		      Unix.setsockopt s Unix.SO_REUSEADDR true;
 		    Unix.bind s (Unix.ADDR_INET(addr,port));
